/*
 * Copyright (C) 2008 Emweb bv, Herent, Belgium.
 *
 * See the LICENSE file for terms of use.
 */

#include "Wt/WHBoxLayout.h"

namespace Wt {

WHBoxLayout::WHBoxLayout()
  : WBoxLayout(LayoutDirection::LeftToRight)
{ }

}
